"""Helper functions for user-supplied arguments and file I/O."""

from __future__ import print_function
import getpass
import os.path
import re
import sys

try:
    from PyPDF2 import PdfFileWriter, PdfFileReader
except ImportError:
    from pyPdf import PdfFileWriter, PdfFileReader


from . import CommandError
import staplelib


ROTATION_NONE = 0
ROTATION_RIGHT = 90
ROTATION_TURN = 180
ROTATION_LEFT = 270
ROTATIONS = {'u': ROTATION_NONE,
             'r': ROTATION_RIGHT,
             'd': ROTATION_TURN,
             'l': ROTATION_LEFT}

HANDLES = {}

def read_pdf(filename):
    """Open a PDF file with PyPDF2."""
    if not os.path.exists(filename):
        raise CommandError("{} does not exist".format(filename))
    pdf = PdfFileReader(open(filename, "rb"))
    if pdf.isEncrypted:
        while True:
            pw = prompt_for_pw(filename)
            matched = pdf.decrypt(pw)
            if matched:
                break
            else:
                print("The password did not match.")
    return pdf


def write_pdf(pdf, filename):
    force = staplelib.OPTIONS.force

    """Write the content of a PdfFileWriter object to a file."""
    if os.path.exists(filename) and not force:
        raise CommandError("File already exists: {}".format(filename))

    opt = staplelib.OPTIONS
    if opt:
        if opt.ownerpw or opt.userpw:
            pdf.encrypt(opt.userpw or '', opt.ownerpw)

    outputStream = open(filename, "wb")
    pdf.write(outputStream)
    outputStream.close()


def prompt_for_pw(filename):
    """Prompt the user for the password to access an input file."""
    print('Please enter a password to decrypt {}.'.format(filename))
    print('(The password will not be shown. Press ^C to cancel).')

    try:
        return getpass.getpass('--> ')
    except KeyboardInterrupt:
        sys.stderr.write('Aborted by user.\n')
        sys.exit(2)


def check_input_files(files):
    """Make sure all input files exist."""

    for filename in files:
        if not os.path.exists(filename):
            raise CommandError("{} does not exist".format(filename))


def check_output_file(filename):
    """Make sure the output file does not exist."""

    if os.path.exists(filename):
        raise CommandError("File already exists: {}".format(filename))


def parse_ranges(handles_files_and_ranges):
    """Parse a list of filenames followed by ranges."""

    operations = []
    handle_pattern = re.compile('^[A-Z]=')
    for inputname in handles_files_and_ranges:
        handle_key = None
        handle_value = None
        if handle_pattern.match(inputname):
            handle_key,handle_value = inputname.split("=",2)
            HANDLES[handle_key] = handle_value
        elif inputname.lower().endswith('.pdf'):
            operations.append({"name": inputname,
                               "pdf": read_pdf(inputname),
                               "pages": []})
        else:
            handle_key = None
            handle_value = None
            match = re.match('([A-Z])?([0-9]+|end)(?:-([0-9]+|end))?([LRD]?)',
                                inputname)
            if not match:
                raise CommandError('Invalid range: {}'.format(inputname))

            if match.group(1):
                handle_key = match.group(1)
                if handle_key in HANDLES:
                    handle_value = HANDLES[handle_key]
                    operations.append({"name": handle_value,
                                       "pdf": read_pdf(handle_value),
                                       "pages": []})
                else:
                    raise CommandError(
                        "Filehandle '{}' does not exist in "
                        "page range '{}'".format(handle_key, inputname))

            current = operations[-1]
            max_page = current['pdf'].getNumPages()
            # allow "end" as alias for the last page
            replace_end = lambda page: (
                max_page if page.lower() == 'end' else int(page))
            begin = replace_end(match.group(2))
            end = replace_end(match.group(3)) if match.group(3) else begin

            rotate = ROTATIONS.get((match.group(4) or 'u').lower())

            if begin > max_page or end > max_page:
                raise CommandError(
                    "Range {}-{} exceeds maximum page number "
                    "{} of file {}".format(
                        begin, end, max_page, current['name']))

            # negative ranges sort pages backwards
            if begin < end:
                pagerange = range(begin, end + 1)
            else:
                pagerange = range(end, begin + 1)[::-1]

            for p in pagerange:
                current['pages'].append((p, rotate))

    return operations



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Main stapler dispatcher."""

from __future__ import print_function

import os
import sys
from argparse import ArgumentParser

import staplelib
from . import commands, CommandError

USAGE = """
usage: %(prog)s [options] mode (input.pdf|input handle) ... [output.pdf]

Modes:
cat/sel: <inputfile>|<input handle> [<pagerange>[<rotation>]] ... (output needed)
    Select the given pages/ranges from input files.
    No range means all pages.
del: <inputfile> [<pagerange>[<rotation>]] ... (output needed)
    Select all but the given pages/ranges from input files.
burst/split: <inputfile> ... (no output needed)
    Create one file per page in input pdf files (no output needed)
zip: <inputfile>|<input handle> [<pagerange>[<rotation>]] ... (output needed)
    Merge/Collate the given input files interleaved.
background: <inputfile> [<pagerange>[<rotation>]] ... (output needed)
    Merge/Overlay the given input files interleaved.
info: <inputfile> ... (no output needed)
    Display PDF metadata
list-log(ical): <inputfile>
    Display the logical names of each page.

Input handle:
    A single, upper-case letter as an alias to a file
    For example: A=input1.pdf B=input2.pdf

Page ranges:
    n - single numbers mean single pages (e.g., 15)
    n-m - page ranges include the entire specified range (e.g. 1-6)
    m-n - negative ranges sort pages backwards (e.g., 6-3)

Extended page range options:
    ...-end will be replaced with the last page in the file
    R, L, or D will rotate the respective range +90, -90, or 180 degrees,
        respectively. (e.g., 1-15R)
""".strip()

# command line option parser
argparser = ArgumentParser(usage=USAGE)
argparser.add_argument('-o', '--ownerpw',
                       action='store',
                       dest='ownerpw',
                       help='Set owner password to encrypt output file with',
                       default=None)
argparser.add_argument('-u', '--userpw',
                       action='store',
                       dest='userpw',
                       help='Set user password to encrypt output file with',
                       default=None)
argparser.add_argument('-v', '--verbose',
                       action='store_true',
                       dest='verbose',
                       default=False)
argparser.add_argument('-f', '--force',
                       action='store_true',
                       dest='force',
                       help='Overwrite output file if it exists',
                       default=False)
argparser.add_argument('-d', '--destdir',
                       dest="destdir",
                       default="." + os.sep,
                       help="directory where to store output file", )
argparser.add_argument('mode',
                       action='store',
                       help="requested stapler mode")


def main(arguments=None):
    """
    Handle all command line arguments and pass them on to the respective
    commands.
    """

    if not arguments:
        arguments = sys.argv[1:]

    (staplelib.OPTIONS, args) = argparser.parse_known_args(args=arguments)

    if not os.path.exists(staplelib.OPTIONS.destdir):
        print_error_and_exit("cannot find output directory named {}".format(
            staplelib.OPTIONS.destdir))

    if (len(args) < 1):
        print_error_and_exit("Not enough arguments", show_usage=True)

    modes = {
        "cat": commands.select,
        "sel": commands.select,
        "split": commands.split,
        "burst": commands.split,
        "del": commands.delete,
        "info": commands.info,
        "zip": commands.zip,
        "background": commands.background,
        "list-log": commands.list_logical_pages,
        "list-logical": commands.list_logical_pages,
    }

    mode = staplelib.OPTIONS.mode

    if mode not in modes:
        print_error_and_exit('Please enter a valid mode', show_usage=True)

    if staplelib.OPTIONS.verbose:
        print("Mode: %s" % mode)

    # dispatch call to known subcommand
    try:
        modes[mode](args)
    except CommandError as e:
        print_error_and_exit(e)


def print_error_and_exit(msg, code=1, show_usage=False):
    """Pretty-print an error to the user."""
    sys.stderr.write(str('Error: {}\n'.format(msg)))

    if show_usage:
        argparser.print_usage(sys.stderr)

    sys.exit(code)



class CommandError(Exception):
    """
    Exception class indicating a problem while executing a stapler command.
    """
    pass

OPTIONS = None # optparse options


def main(arguments=None):
    from . import stapler
    stapler.main(arguments)



#!/usr/bin/env python

import os
import shutil
import tempfile
import unittest

from PyPDF2.pdf import PdfFileReader

from staplelib import main, CommandError

HERE = os.path.abspath(os.path.dirname(__file__))
TESTFILE_DIR = os.path.join(HERE, 'testfiles')
STAPLER = os.path.join(HERE, '..', 'stapler')
ONEPAGE_PDF = os.path.join(TESTFILE_DIR, '1page.pdf')
FIVEPAGE_PDF = os.path.join(TESTFILE_DIR, '5page.pdf')


def run_stapler(arguments):
    main(arguments)


class TestStapler(unittest.TestCase):
    """Some unit tests for the stapler tool."""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.outputfile = os.path.join(self.tmpdir, 'output.pdf')
        os.chdir(self.tmpdir)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        os.chdir(HERE)

    def test_cat(self):
        """Make sure files are properly concatenated."""
        run_stapler(['cat', ONEPAGE_PDF, FIVEPAGE_PDF,
                     self.outputfile])
        self.assertTrue(os.path.isfile(self.outputfile))
        with open(self.outputfile, 'rb') as outputfile:
            pdf = PdfFileReader(outputfile)
            self.assertEqual(pdf.getNumPages(), 6)

    def test_sel_one_page(self):
        """Test select of a one page from a PDF file."""
        run_stapler(['sel', 'A=' + FIVEPAGE_PDF, 'A2', self.outputfile])
        self.assertTrue(os.path.isfile(self.outputfile))
        with open(self.outputfile, 'rb') as outputfile:
            pdf = PdfFileReader(outputfile)
            self.assertEqual(pdf.getNumPages(), 1)

    def test_sel_range(self):
        """Test select of more pages from a PDF file."""
        run_stapler(['cat', 'A=' + FIVEPAGE_PDF, 'A2-4', self.outputfile])
        self.assertTrue(os.path.isfile(self.outputfile))
        with open(self.outputfile, 'rb') as outputfile:
            pdf = PdfFileReader(outputfile)
            self.assertEqual(pdf.getNumPages(), 3)

    def test_del_one_page(self):
        """Test del command for inverse select of one page."""
        run_stapler(['del', 'A=' + FIVEPAGE_PDF, 'A1', self.outputfile])
        self.assertTrue(os.path.isfile(self.outputfile))
        with open(self.outputfile, 'rb') as outputfile:
            pdf = PdfFileReader(outputfile)
            self.assertEqual(pdf.getNumPages(), 4)

    def test_del_range(self):
        """Test del command for inverse select multiple pages."""
        run_stapler(['del', 'A=' + FIVEPAGE_PDF, 'A2-4', self.outputfile])
        self.assertTrue(os.path.isfile(self.outputfile))
        with open(self.outputfile, 'rb') as outputfile:
            pdf = PdfFileReader(outputfile)
            self.assertEqual(pdf.getNumPages(), 2)

    def test_split(self):
        """Make sure a file is properly split into pages."""
        run_stapler(['split', FIVEPAGE_PDF])

        filelist = os.listdir(self.tmpdir)
        self.assertEqual(len(filelist), 5)
        for f in os.listdir(self.tmpdir):
            with open(os.path.join(self.tmpdir, f), 'rb') as pdf_file:
                pdf = PdfFileReader(pdf_file)
                self.assertEqual(pdf.getNumPages(), 1)

    def test_zip(self):
        """Test zip."""
        run_stapler(['zip', ONEPAGE_PDF, FIVEPAGE_PDF, self.outputfile])
        self.assertTrue(os.path.isfile(self.outputfile))
        with open(self.outputfile, 'rb') as outputfile:
            pdf = PdfFileReader(outputfile)
            self.assertEqual(pdf.getNumPages(), 6)

    def test_output_file_already_exists(self):
        """Test zip."""
        with self.assertRaises(SystemExit) as e:
            run_stapler(['zip', ONEPAGE_PDF, FIVEPAGE_PDF])
        self.assertEqual(e.exception.code, 1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['staplelib']

package_data = \
{'': ['*'], 'staplelib': ['testfiles/*']}

install_requires = \
['PyPDF2>=1.26', 'more-itertools>=2.2,<6.0.0']

entry_points = \
{'console_scripts': ['pdf-stapler = staplelib:main',
                     'stapler = staplelib:main']}

setup_kwargs = {
    'name': 'stapler',
    'version': '1.0.0',
    'description': 'Manipulate PDF documents from the command line',
    'long_description': "Stapler\n=======\n\nStapler is a pure Python alternative to\n`PDFtk <http://www.pdfhacks.com/pdftk/>`__, a tool for manipulating PDF\ndocuments from the command line.\n\nHistory\n-------\n\nPDFtk was written in Java and C++, and is natively compiled with gcj.\nSadly, it has been discontinued a few years ago and bitrot is setting in\n(e.g., it does not compile easily on a number of platforms).\n\nPhilip Stark decided to look for an alternative and found pypdf, a PDF\nlibrary written in pure Python. He couldn't find a tool which actually\nused the library, so he started writing his own.\n\nVersion 0.3 of stapler was completely refactored by Fred Wenzel. He also\nadded tests and awesome functionality.\n\nLike pdftk, stapler is a command-line tool. If you would like to add a\nGUI, compile it into a binary for your favorite platform, or contribute\nanything else, feel free to fork and send a pull request.\n\nContributors and Authorship\n---------------------------\n\nStapler version 0.2 was written in 2009 by Philip Stark. Stapler version\n0.3 was written in 2010 by Fred Wenzel.\n\nFor a list of contributors, check the ``CONTRIBUTORS`` file.\n\nChange log (sorta)\n------------------\n\n- **1.0.0** Port to Python 3. Replace OptionParser with more\n  modern ArgumentParser. Cleaning up repository.\n\n- **0.3.3** include try-except blocks for supporting legacy pyPdf\n  if needed. Also fixes some PyPI issues like the missing License Trove\n  classifier and some dependencies.\n\n- **0.3.0** Refactoring by Fred Wenzel and now using PyPDF2\n\n- **0.2.0** Feature completeness using original pyPdf\n\nLicense\n-------\n\nStapler is distributed under a BSD license. A copy of the BSD Style\nLicense used can be found in the file ``LICENSE``.\n\nUsage\n-----\n\nThere are the following modes in Stapler:\n\nselect/delete (called with ``sel`` and ``del``, respectively)\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n``sel`` is also available as ``cat`` for compatibility with my\npersonal muscle memory. :)\n\nWith select, you can cherry-pick pages from pdfs and concatenate them\ninto a new pdf file.\n\nInput files can be associated with handles for use with ranges later.\nA handle is a single, upper-case letter:\n\n::\n\n    <input handle>=<input>\n\nSyntax:\n\n::\n\n    stapler sel input1 page_or_range [page_or_range ...] [input2 p_o_r ...]\n\nExamples:\n\n::\n\n    # concatenate a and b into output.pdf\n    stapler sel a.pdf b.pdf output.pdf\n\n    # generate a pdf file called output.pdf with the following pages:\n    # 1, 4-8 in 180° (D for down), 20-40 from a.pdf, 1-5 from b.pdf in \n    # this order\n    stapler sel a.pdf 1 4-8D 20-40 b.pdf 1-5 output.pdf\n\n    # the same example with a handle for b.pdf\n    stapler sel B=b.pdf a.pdf 1 4-8D 20-40 B1-5 output.pdf\n\n    # generate a pdf file called output.pdf with the following pages:\n    # 1 from a.pdf, 1-5 from b.pdf, 4-8 in 180° (D for down), 20-40 from a.pdf\n    # this order\n    stapler sel A=a.pdf B=b.pdf A1 B1-5 A4-8D A20-40 output.pdf\n\n    # reverse some of the pages in a.pdf by specifying a negative range\n    stapler sel a.pdf 1-3 9-6 10 output.pdf\n\nThe delete command works almost exactly the same as select, but inverse.\nIt uses the pages and ranges which you *didn't* specify.\n\nsplit/burst:\n~~~~~~~~~~~~\n\nSplits the specified pdf files into their single pages and writes each\npage into it's own pdf file with this naming scheme:\n\n::\n\n    ${origname}_${zero-padded page no}.pdf\n\nSyntax:\n\n::\n\n    stapler split input1 [input2 input3 ...]\n\nExample for a file foobar.pdf with 20 pages:\n\n::\n\n    $ stapler split foobar.pdf\n    $ ls\n    foobar_01.pdf foobar_02.pdf ... foobar_19.pdf foobar_20.pdf\n\nMultiple files can be specified, they will be processed as if you called\nsingle instances of stapler.\n\nzip:\n~~~~\n\nWith zip, you can cherry-pick pages from pdfs (like select). The pages\nfrom each pdf are merged together in an interleaving manner. This can be\nused to collate a pdf with odd pages and a pdf with even pages into a\nsingle file.\n\nSyntax: stapler zip input1 [range[rotation]] [range ...] [input2\n[range...] ...] out\n\nExamples:\n\n::\n\n    # combine a pdf with odd pages and a pdf with even pages into output.pdf\n    stapler zip odd.pdf even.pdf output.pdf\n\n    # combine a.pdf b.pdf and c.pdf, but use only some pages of c.pdf and\n    #  rotate b.pdf right (90° clockwise) and rotate c.pdf left (90° counter-\n    # clockwise)\n    stapler zip a.pdf b.pdf 1-endR c.pdf 1-3L output.pdf\n\nIf one of the ranges is shorter than the others, stapler will continue\nto merge the remaining pages.\n\ninfo:\n~~~~~\n\nShows information on the metadata stored inside a PDF file.\n\nSyntax:\n\n::\n\n    stapler info foo.pdf\n\nExample output:\n\n::\n\n    \\*\\*\\* Metadata for foo.pdf\n\n    /ModDate:  D:20100313082451+01'00'\n    /CreationDate:  D:20100313082451+01'00'\n    /Producer:  GPL Ghostscript 8.70\n    /Title:  foo.pdf\n    /Creator:  PDFCreator Version 0.9.9\n    /Keywords:\n    /Author:  John Doe\n    /Subject:\n\nlist-logical:\n~~~~~~~~~~~~~\n\nShows each logical page number and the associated physical page number.\n\nSyntax:\n\n::\n\n    stapler list-logical foo.pdf\n\nExample output:\n\n::\n\n    A-1\t1\n    C-1\t2\n    D-1\t3\n    D-2\t4\n    D-3\t5\n    D-4\t6\n",
    'author': 'Philip Stark',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)



"""Module containing the actual commands stapler understands."""
from __future__ import print_function
import math
import os

try:
    from PyPDF2 import PdfFileWriter, PdfFileReader
except:
    from pyPdf import PdfFileWriter, PdfFileReader
import itertools
import more_itertools

from . import CommandError, iohelper
import staplelib


def select(args, inverse=False):
    """
    Concatenate files / select pages from files.

    inverse=True excludes rather than includes the selected pages from
    the file.
    """

    filesandranges = iohelper.parse_ranges(args[:-1])
    outputfilename = args[-1]
    verbose = staplelib.OPTIONS.verbose

    if not filesandranges or not outputfilename:
        raise CommandError("Both input and output filenames are required.")

    output = PdfFileWriter()
    try:
        for input in filesandranges:
            pdf = input['pdf']
            if verbose:
                print(input['name'])

            # empty range means "include all pages"
            if not inverse:
                pagerange = input['pages'] or [
                    (p, iohelper.ROTATION_NONE) for p in
                    range(1, pdf.getNumPages() + 1)]
            else:
                excluded = [p for p, r in input['pages']]
                pagerange = [(p, iohelper.ROTATION_NONE) for p in
                             range(1, pdf.getNumPages() + 1) if
                             p not in excluded]

            for pageno, rotate in pagerange:
                if 1 <= pageno <= pdf.getNumPages():
                    if verbose:
                        print("Using page: {} (rotation: {} deg.)".format(
                            pageno, rotate))

                    output.addPage(pdf.getPage(pageno-1)
                                   .rotateClockwise(rotate))
                else:
                    raise CommandError("Page {} not found in {}.".format(
                        pageno, input['name']))

    except Exception as e:
        raise CommandError(e)

    if os.path.isabs(outputfilename):
        iohelper.write_pdf(output, outputfilename)
    else:
        iohelper.write_pdf(output, staplelib.OPTIONS.destdir +
                           os.sep + outputfilename)


def delete(args):
    """Concatenate files and remove pages from files."""
    return select(args, inverse=True)


def split(args):
    """Burst an input file into one file per page."""
    files = args
    verbose = staplelib.OPTIONS.verbose

    if not files:
        raise CommandError("No input files specified.")

    inputs = []
    try:
        for f in files:
            inputs.append(iohelper.read_pdf(f))
    except Exception as e:
        raise CommandError(e)

    filecount = 0
    pagecount = 0
    for input in inputs:
        # zero-padded output file name
        (base, ext) = os.path.splitext(os.path.basename(files[filecount]))
        output_template = ''.join([
            base,
            '_',
            '%0',
            str(math.ceil(math.log10(input.getNumPages()))),
            'd',
            ext
        ])

        for pageno in range(input.getNumPages()):
            output = PdfFileWriter()
            output.addPage(input.getPage(pageno))

            outputname = output_template % (pageno + 1)
            if verbose:
                print(outputname)
            iohelper.write_pdf(output, staplelib.OPTIONS.destdir +
                               os.sep + outputname)
            pagecount += 1
        filecount += 1

    if verbose:
        print("\n{} page(s) in {} file(s) processed.".format(
            pagecount, filecount))


def info(args):
    """Display Metadata content for all input files."""
    files = args

    if not files:
        raise CommandError("No input files specified.")

    for f in files:
        pdf = iohelper.read_pdf(f)
        print("*** Metadata for {}".format(f))
        print()
        info = pdf.documentInfo
        if info:
            for name, value in info.items():
                print(u"    {}:  {}".format(name, value))
        else:
            print("    (No metadata found.)")
        print()

def background(args):
    """Combine 2 files with corresponding pages merged."""

    filesandranges = iohelper.parse_ranges(args[:-1])
    outputfilename = args[-1]
    verbose = staplelib.OPTIONS.verbose

    if not filesandranges or not outputfilename:
        raise CommandError("Both input and output filenames are required.")

    try:
        filestozip = []
        for input in filesandranges:
            pdf = input['pdf']
            if verbose:
                print(input['name'])

            # empty range means "include all pages"
            pagerange = input['pages'] or [
                (p, iohelper.ROTATION_NONE) for p in
                range(1, pdf.getNumPages() + 1)]

            pagestozip = []
            for pageno, rotate in pagerange:
                if 1 <= pageno <= pdf.getNumPages():
                    if verbose:
                        print("Using page: {} (rotation: {} deg.)".format(
                            pageno, rotate))

                    pagestozip.append(pdf.getPage(pageno-1)
                                   .rotateClockwise(rotate))
                else:
                    raise CommandError("Page {} not found in {}.".format(
                        pageno, input['name']))
            filestozip.append(pagestozip)

        output = PdfFileWriter()
        for pagelist in list(itertools.izip_longest(*filestozip)):
            page = pagelist[0]
            for p in pagelist[1:]:
              if not page:
                page = p
              elif p:
                page.mergePage(p)
            output.addPage(page)

    except Exception as e:
        import sys
        import traceback
        traceback.print_tb(sys.exc_info()[2])
        raise CommandError(e)

    if os.path.isabs(outputfilename):
        iohelper.write_pdf(output, outputfilename)
    else:
        iohelper.write_pdf(output, staplelib.OPTIONS.destdir +
                           os.sep + outputfilename)

def zip(args):
    """Combine 2 files with interleaved pages."""
    filesandranges = iohelper.parse_ranges(args[:-1])
    outputfilename = args[-1]
    verbose = staplelib.OPTIONS.verbose

    if not filesandranges or not outputfilename:
        raise CommandError('Both input and output filenames are required.')

    # Make [[file1_p1, file1_p2], [file2_p1, file2_p2], ...].
    filestozip = []
    for input in filesandranges:
        pdf = input['pdf']
        if verbose:
            print(input['name'])

        # Empty range means "include all pages".
        pagerange = input['pages'] or [
            (p, iohelper.ROTATION_NONE) for p in
            range(1, pdf.getNumPages() + 1)]

        pagestozip = []
        for pageno, rotate in pagerange:
            if 1 <= pageno <= pdf.getNumPages():
                if verbose:
                    print("Using page: {} (rotation: {} deg.)".format(
                        pageno, rotate))

                pagestozip.append(
                    pdf.getPage(pageno - 1).rotateClockwise(rotate))
            else:
                raise CommandError("Page {} not found in {}.".format(
                    pageno, input['name']))
        filestozip.append(pagestozip)

    # Interweave pages.
    output = PdfFileWriter()
    for page in more_itertools.roundrobin(*filestozip):
        output.addPage(page)

    if os.path.isabs(outputfilename):
        iohelper.write_pdf(output, outputfilename)
    else:
        iohelper.write_pdf(output, staplelib.OPTIONS.destdir +
                           os.sep + outputfilename)


def int_to_page_alpha(pageno, base):
    """return uppercase alphabetic page numbers for PAGENO starting at BASE (a or A).
Adobe defines them as A to Z, then AA to ZZ, and so on.
Yes, that is somewhat wacky."""
    (div, mod) = divmod( pageno-1, 26)
    c = chr(mod + ord(base))
    return c * (div+1)

# next text is from Paul M. Winkler
# via https://www.oreilly.com/library/view/python-cookbook/0596001673/ch03s24.html
def int_to_roman(input):
    """ Convert an integer to a Roman numeral. """

    if not isinstance(input, type(1)):
        raise TypeError("expected integer, got %s" % type(input))
    if not 0 < input < 4000:
        raise ValueError("Argument must be between 1 and 3999")
    ints = (1000, 900,  500, 400, 100,  90, 50,  40, 10,  9,   5,  4,   1)
    nums = ('M',  'CM', 'D', 'CD','C', 'XC','L','XL','X','IX','V','IV','I')
    result = []
    for i in range(len(ints)):
        count = int(input / ints[i])
        result.append(nums[i] * count)
        input -= ints[i] * count
    return ''.join(result)



#
# pdf_page_enumeration is
# inspired by  https://stackoverflow.com/questions/12360999/retrieve-page-numbers-from-document-with-pypdf
# (thanks vjayky!)
# and informed by https://www.w3.org/TR/WCAG20-TECHS/PDF17.html
# (thanks, w3c!)
# which recaps the PDF-1.7 specification
# https://www.adobe.com/content/dam/acom/en/devnet/pdf/pdfs/PDF32000_2008.pdf
#
def pdf_page_enumeration(pdf):
    """Generate a list of pages, using /PageLabels (if it exists).  Returns a list of labels."""
    
    try:
        pagelabels = pdf.trailer["/Root"]["/PageLabels"]
    except:
        # ("No /Root/PageLabels object"), so infer the list.
        return range(1, pdf.getNumPages() + 1)
    
    # """Select the item that is most likely to contain the information you desire; e.g.
    #        {'/Nums': [0, IndirectObject(42, 0)]}
    #    here, we only have "/Num". """
    
    try:
        pagelabels_nums = pdf.trailer["/Root"]["/PageLabels"]["/Nums"]
    except:
        raise CommandError("Malformed PDF, /Root/PageLabels but no .../Nums object")

    #
    # At this point we have either the object or the list.
    # Make it a list.
    #
    if isinstance(pagelabels_nums, (list,)):
        pagelabels_nums_list = pagelabels_nums
    else:
        pagelabels_nums_list = list(pagelabels_nums)

    labels = []
    style = None
    # default
    style = '/D'
    prefix = ''
    next_pageno = 1
    for i in range(0, pdf.getNumPages()):
        if len(pagelabels_nums_list) > 0 and i >= pagelabels_nums_list[0]:
            pagelabels_nums_list.pop(0)  # discard index
            pnle = pagelabels_nums_list.pop(0)
            style = pnle.get('/S', '/D')
            prefix = pnle.get('/P', '')
            next_pageno = pnle.get('/St', 1)
        pageno_str = ''
        if style == '/D':
            pageno_str = str(next_pageno)
        elif style == '/A':
            pageno_str = int_to_page_alpha(next_pageno, 'A')
        elif style == '/a':
            pageno_str = int_to_page_alpha(next_pageno, 'a')
        elif style == '/R':
            pageno_str = int_to_roman(next_pageno)
        elif style == '/r':
            pageno_str = int_to_roman(next_pageno).lower()
        else:
            raise CommandError("Malformded PDF: unkown page numbering style " + style)
        labels.append(prefix + pageno_str)
        next_pageno += 1

    return labels


def list_logical_pages(args):
    """List the logical names of each page."""
    verbose = staplelib.OPTIONS.verbose

    files = args
    if not files:
        raise CommandError("An input filename is required.")

    try:
        for input in files:
            pdf = iohelper.read_pdf(input)
            if verbose:
                print(input)
            i = 0
            for label in pdf_page_enumeration(pdf):
                i += 1
                print("{}\t{}".format(label, str(i)))

    except Exception as e:
        raise CommandError(e)
